/*
 * Common file resource classes.
 *
 * Copyright (C) 2018, 2019, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#include <fsclient/fsdesc.h>
#include <fsserver/accessor.h>
#include <fsserver/pager.h>
#include <fsserver/resource.h>
#include <fsserver/sync_support.h>

/* Common interfaces. */

#include <fsserver/mapped_file_object_interface.h>



/* A file resource base class providing common implementations. */

class FileResourceBase : public Resource, public SyncSupport
{
protected:
  /* Access to data. */

  FileAccessor *_accessor;

  /* Shared dataspace details. */

  fs_object_t _obj;

  void reset()
  {
    _obj.start_pos = 0;
    _obj.end_pos = 0;
  }

public:
  explicit FileResourceBase(FileAccessor *accessor, fs_object_t obj)
  : Resource(), SyncSupport(), _accessor(accessor), _obj(obj)
  {
    reset();
  }

  /* Deallocation of resources. */

  virtual void close();

  /* Activation, ensuring that deallocation will not occur concurrently. */

  virtual void activate()
  { _accessor->activate(); }

  /* Operation methods. */

  virtual long begin(size_t *size);

  virtual long begin_first(size_t *size);

  virtual long flush(size_t data_start, size_t *size);

  virtual long fstat(size_t *size, l4_cap_idx_t *statcap);

  virtual long yield(size_t *size);

  /* Queue access. */

  virtual Queue *queue()
  { return _accessor->queue(); }

  /* Size access. */

  virtual size_t get_size()
  { return _accessor->get_size(); }
};



/* A memory-mapped file resource. */

class MappedFileResource : public FileResourceBase, public Pager,
                           public MappedFileObject
{
public:
  explicit MappedFileResource(PagingAccessor *accessor, fs_object_t obj)
  : FileResourceBase(accessor, obj)
  { 
  }

  /* Server details. */

  int expected_items();

  ipc_server_handler_type handler();

  void *interface()
  { return static_cast<MappedFileObject *>(this); }

  /* Pager methods. */

  virtual PagingAccessor *get_accessor()
  { return dynamic_cast<PagingAccessor *>(_accessor); }

  virtual unsigned long get_data_start()
  { return _obj.start_pos; };

  virtual unsigned long get_data_size()
  { return _obj.size; };

  /* Deallocation of resources. */

  virtual void close();

  /* File and synchronisation operations. */

  virtual long begin(size_t *size)
  { return FileResourceBase::begin(size); }

  virtual long begin_first(size_t *size)
  { return FileResourceBase::begin_first(size); }

  virtual long end()
  { return FileResourceBase::end(); }

  virtual long flush(size_t data_start, size_t *size);

  virtual long fstat(size_t *size, l4_cap_idx_t *statcap)
  { return FileResourceBase::fstat(size, statcap); }

  virtual long yield(size_t *size)
  { return FileResourceBase::yield(size); }

  /* Mapped file operations. */

  virtual long mmap(size_t position, size_t length, size_t *start_pos,
                    size_t *end_pos, size_t *data_end);

  /* Dataspace operations. */

  virtual long map(unsigned long offset, l4_addr_t hot_spot,
                   unsigned long flags, l4_snd_fpage_t *region)
  { return Pager::map(offset, hot_spot, flags, region); }
};
