/*
 * Accessor object functionality.
 *
 * Copyright (C) 2019, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/c/util/cap_alloc.h>
#include <l4/re/env.h>
#include <l4/sys/factory.h>
#include <l4/sys/types.h>

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include "accessor.h"

/* Obtain declarations not obtained in the accessor header. */

#include "pages.h"
#include "paging.h"



/* Initialisation. */

Accessor::Accessor()
{
  _accessor_semaphore = l4re_util_cap_alloc();

  if (l4_is_valid_cap(_accessor_semaphore))
  {
    l4_factory_create(l4re_env()->factory, L4_PROTO_SEMAPHORE, _accessor_semaphore);
    unlock_accessor();
  }

  _queue.start();
}

Accessor::~Accessor()
{
  if (l4_is_valid_cap(_accessor_semaphore))
    l4re_util_cap_free_um(_accessor_semaphore);

  _queue.end();
}



/* Return the common page collection. */

Pages *PagingAccessor::pages()
{
  return _paging->pages();
}



/* Accounting methods. */

void PagingAccessor::finalise()
{
  printf("Finalise accessor: %p\n", this);

  _paging->remove(get_fileid());

  _map.purge(this, pages());

  printf("Finalised accessor: %p\n", this);
}



/* Flexpage storage methods. */

void PagingAccessor::fill_flexpage_data(Flexpage *flexpage)
{
  /* Update the data length according to the current size. */

  flexpage->set_data_size(_size);

  read(flexpage->get_data_start(),
       (char *) flexpage->get_base(),
       flexpage->get_data_length());

  /* Pad the flexpage where the data will not fill it (such as at the end of
     the file). */

  flexpage->pad();
}

void PagingAccessor::flush_flexpage_data(Flexpage *flexpage)
{
  /* Update the data length according to the current size. */

  flexpage->set_data_size(_size);

  write(flexpage->get_data_start(),
        (char *) flexpage->get_base(),
        flexpage->get_data_length());
}



/* Get or create a flexpage for the given details. This is the exposed method
   for obtaining flexpages, serialising access in order to prevent successive
   requests obtaining different results for the same inputs. */

Flexpage *PagingAccessor::get_flexpage(unsigned long data_start,
                                       unsigned long data_size,
                                       unsigned long offset,
                                       l4_addr_t hot_spot,
                                       unsigned long flags)
{
  lock_accessor();

  /* Find the flexpage or obtain a new one. */

  Flexpage *flexpage = find_flexpage(data_start, offset, flags);

  if ((flexpage != NULL) && (flexpage->valid()))
  {
    /* Upgrade any incompatible flexpage. */

    if (!flexpage->supports_flags(flags))
      flexpage->upgrade(flags);
  }
  else
    flexpage = new_flexpage(data_start, data_size, offset, hot_spot, flags);

  printf("return flexpage %p\n", flexpage);
  return flexpage;
}



/* Queue an issued flexpage after having sent it. This is the exposed method
   for finalising issued flexpages. */

void PagingAccessor::queue_flexpage(Flexpage *flexpage)
{
  /* Increment the reference count of flexpage users. */

  flexpage->increment();

  /* Finally, make the flexpage available for eventual reuse. */

  pages()->queue_flexpage(this, flexpage);

  unlock_accessor();
}



/* Allow the page collection to remove a flexpage. */

void PagingAccessor::remove_flexpage(Flexpage *flexpage)
{
  if (_map.remove(flexpage))
    flush_flexpage(flexpage, false, true);
}



/* Find a flexpage for the given details. Any found flexpage is tested for
   current availability by removing it from the available queue, if possible.
   If it is unavailable, it is subject to deletion by another accessor and is
   disregarded. */

Flexpage *PagingAccessor::find_flexpage(unsigned long data_position,
                                        unsigned long offset, unsigned long flags)
{
  printf("find_flexpage(%ld, %ld, %lx)\n", data_position, offset, flags);
  Flexpage *flexpage = _map.find(data_position + offset);

  /* Attempt to make the flexpage unavailable for others. If the flexpage is
     not available, disregard it. */

  if ((flexpage != NULL) && flexpage->valid() && pages()->reserve_flexpage(this, flexpage))
    return flexpage;
  else
    return NULL;
}



/* Set a new flexpage in the page collection for the given details, obtaining
   more memory or reclaiming memory from existing flexpages that can be retired
   by flushing their data. */

Flexpage *PagingAccessor::new_flexpage(unsigned long data_start,
                                       unsigned long data_size,
                                       unsigned long offset,
                                       l4_addr_t hot_spot,
                                       unsigned long flags)
{
  /* Obtain more memory if possible. */

  Flexpage *flexpage = pages()->new_flexpage();

  /* Recycle flexpages when reaching the memory limit. */

  if (flexpage == NULL)
  {
    /* Obtain the next flexpage for reuse, with it being flushed and removed from
       its current accessor. */

    flexpage = pages()->remove_flexpage();
  }

  /* Reset (and recycle) the flexpage. */

  flexpage->reset(offset, hot_spot, flags);

  /* Associate new flexpages with file regions. */

  flexpage->set_data_region(data_start, data_size);
  fill_flexpage_data(flexpage);

  /* Register the flexpage in this accessor. */

  _map.insert(flexpage);

  return flexpage;
}



/* Flush all allocated flexpages within a given region, writing data to
   storage. If the keep indicator is set to a true value, the flexpages will be
   retained; otherwise, they are invalidated. */

void PagingAccessor::flush_all(unsigned long start, unsigned long end, bool keep)
{
  _map.flush_all(start, end, keep, this, pages());
}



/* Flush the given flexpage, copying data from the flexpage using the
   flush_flexpage_data method provided by derived classes. If the keep indicator
   is set to a true value, the flexpage will be retained; otherwise, the
   flexpage is invalidated. */

void PagingAccessor::flush_flexpage(Flexpage *flexpage, bool keep, bool purge)
{
  if (flexpage->valid())
  {
    /* Only flush writable flexpages. */

    if (flexpage->writable())
      flush_flexpage_data(flexpage);

    /* Only flush flexpages where no other users are involved. */

    if (purge || (!keep && !flexpage->decrement()))
    {
      flexpage->unmap();
      flexpage->invalidate();
    }
  }
}
