/*
 * Pipe resource functionality.
 *
 * Copyright (C) 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <stdio.h>

#include <ipc/server.h>

#include "pipe_object_server.h"
#include "pipe_resource.h"
#include "resource_server.h"



/* Resource-closing methods. */

void PipeResource::close()
{
  /* Flush any open pages. */

  flush_all();

  /* Detach from the accessor, freeing it if no object is using it. */

  PipePagingAccessor *accessor = _accessor, *next;

  if (_writer)
  {
    if (!accessor->detach())
      delete accessor;
  }
  else
  {
    while (accessor)
    {
      next = accessor->next_accessor();

      if (!accessor->detach())
        delete accessor;

      accessor = next;
    }
  }

  /* NOTE: For debugging. */

  printf("Closed pipe.\n");
}



/* Server details. */

int PipeResource::expected_items()
{
  return PipeObject_expected_items;
}

ipc_server_handler_type PipeResource::handler()
{
  return (ipc_server_handler_type) handle_PipeObject;
}

/* Pipe resource operations. */

long PipeResource::flush(size_t data_start, size_t *size)
{
  flush_all(true);

  /* Update the extent of data in the accessor. Since each accessor effectively
     maintains access to a kind of file in its entirety, data_start will advance
     from zero up to the amount of data written. */

  if (_writer && (data_start > _accessor->get_size()))
    _accessor->set_size(data_start);

  *size = _accessor->get_size();
  return L4_EOK;
}

long PipeResource::current_region(size_t *size)
{
  *size = _size;
  return L4_EOK;
}

long PipeResource::next_region(size_t *size, size_t *data_end)
{
  /* Obtain a new accessor if writing. */

  if (_writer)
  {
    /* Assume that the region is now full. */

    _accessor->set_size(_size);

    /* Release the accessor and acquire another. */

    _accessor->detach();
    _accessor = _accessor->new_accessor();
  }

  /* Flush and unmap all current flexpages if reading since this means that they
     are no longer needed. */

  else
  {
    flush_all(false);

    /* Obtain and reference any next accessor. */

    PipePagingAccessor *accessor = _accessor->next_accessor();

    /* Without a subsequent accessor, indicate failure. */

    if (!accessor)
      return -L4_EIO;

    /* Detach for both the reader and the writer. */

    if (!_accessor->detach())
      delete _accessor;

    _accessor = accessor;
  }

  *size = _size;
  *data_end = _accessor->get_size();
  return L4_EOK;
}
