/*
 * Common server functionality.
 *
 * Copyright (C) 2018, 2019 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <ipc/server.h>
#include "server.h"



/* Server methods. */

long Server::bind(const char *name)
{
  return server_bind(name, this);
}

long Server::start()
{
  return server_loop(expected_items());
}



/* Server binding/registration. */

long server_bind(const char *name, Server *obj)
{
  l4_cap_idx_t server;

  return ipc_server_bind(name, l4_umword_t(obj), &server);
}



/* Common dispatch functionality. */

void server_handle(ipc_message_t *msg, Server *obj)
{
  /* Capture the message details, dispatch and reply. */

  ipc_message_open(msg);
  obj->dispatch(msg);

  /* Discard any propagated capabilities before waiting. */

  ipc_message_discard(msg);
}



/* A general server main loop. */

long server_loop(int expected_items)
{
  /* The target server object is determined from each message using the
     label. */

  Server *target;
  ipc_message_t msg;
  l4_umword_t label;

  /* Declare the extent to which capabilities are expected in messages. */

  long err = ipc_message_expect(&msg, expected_items);
  if (err)
    return err;

  /* Wait for an incoming message. */

  while (1)
  {
    ipc_message_wait(&msg, &label);

    /* Ignore erroneous messages. */

    if (l4_ipc_error(msg.tag, l4_utcb()))
      continue;

    /* Handle messages for different server objects. */

    target = reinterpret_cast<Server *>(label & ~3UL);

    /* NOTE: Threads could be employed when dispatching. */

    server_handle(&msg, target);
  }

  return L4_EOK;
}
